
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		      Lineprinter Queueing Process (PRINT)
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	21 Sep 82  DPK	Modified to use ftime(2) and getdate(3).
 *
 *	15 Nov 82  DPK	Added fclose() in filecopying loop.
 *
 *	18 Nov 82  DPK	Check for no printable files added.
 *
 *	10 Dec 82  DPK	Modified handling of forms variable.
 *
 *	20 Dec 82  DPK	Elinimated the fatal() function.
 *
 *	13 Apr 83  DPK  Moved several function out into library.
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>
#include "queue.h"
#include <sys/stat.h>
#include <sys/timeb.h>

extern	int	Cp_min, Cp_max;
extern	int	Pri_min, Pri_max;
extern	char	*Tmpqdir;
extern	char	*Formsfile;

/*
 *	Configurations Values for PRINT
 */
int	Printpri	= 64;		/* Default priority, (in the middle) */
char	*Printforms	= "white";	/* Default forms for LPR */
char	*Printqueue	= "lp";		/* Default queue for printing */
char	*Sysmgr		= "mdqs";	/* Default orphanage for messages */
char	*Hostname	= "localhost";	/* Default local host name */
char	*Headerdir	= "/usr/lib/q/headers";
					/* Location of "standard" headers */
char	*Printhdr	= "/usr/lib/q/headers/default";
					/* Default lp header */

char	*header		= NULL;
char	*forms		= NULL;
char	*queue		= NULL;
char	*title		= NULL;		/* Job title */
char	*username	= NULL;		/* Internet Return Address */
char	missingfmt[] = "Missing %s argument";
char	dupfmt[] = "Duplicate %s argument given";
char	newlinefmt[] = "Illegal newline in %s argument";

int	copies = 0;
int	priority = 0;
int	lpoptions = 0;
int	qoptions = 0;
int	verbose = 0;
int	error = 0;
int 	oneheader= 1;
time_t	atime = 0L;
int	nfiles = 0;
int	nfails = 0;
char	dataname[QDIRSIZ];

extern FILE	*mkdata();
extern FILE	*mkcontrol();
extern char	*ckforms();
extern char	*rindex();
extern char	*strdup();
extern char	**findque();
extern time_t	getdate();
extern struct	passwd *getpwuid();
extern int	cleanup();
extern long	copy();

struct	request r;
time_t	timenow;
FILE	*cf;			/* The control file */
FILE	*df;			/* A data file */


struct paramtab paramtab[] = {
	"sysmgr",	STRPVAR,	&Sysmgr,		NULL,
	"hostname",	STRPVAR,	&Hostname,		NULL,
	"print-forms",	STRPVAR,	&Printforms,		NULL,
	"print-queue",	STRPVAR,	&Printqueue,		NULL,
	"print-hdr",	STRPVAR,	&Printhdr,		NULL,
	"print-hdrdir",	STRPVAR,	&Headerdir,		NULL,
	"print-prior",	INTVAR,		(char **) &Printpri,	NULL,
	0,		0,		0,			0
};

/*  Storage for pointer to Xtended arguments (-x thing=foo)  */
#define	NXARGS	50
char	*xargs[NXARGS];
int	xargcnt = 0;

/*
 *		M A I N ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	FILE	*fin;			/* The input file to be printed */
	struct	passwd *pw;
	char	*arg;
	char	**qp;
	int	i;

	rdconf( paramtab );
	time( &timenow );
	r.r_gid = getgid();
	r.r_uid = getuid();
	r.r_copies = 1;
	r.r_pri = Printpri;
	r.r_tstart = 0L;

	signal( SIGHUP, cleanup );
	if (signal(SIGQUIT, SIG_IGN) != SIG_IGN)
	 	(void)	signal( SIGQUIT, cleanup );
	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
		(void)	signal( SIGINT, cleanup );

	for (i = 1; i < argc; i++) {
		arg = argv[i];
		if (*arg != '-') {
			continue;
		}
		arg++;
		while (*arg) {
			switch (*arg) {
			case 'a':	/* Run after time */
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "time");
					error++;
					break;
				} else
					i++;
				if (atime != 0L) {
					err (dupfmt, "time");
					error++;
				} else if ((atime = getdate(argv[i], NULL)) == -1L) {
					err ("Bad time specification");
					error++;
				} else
					r.r_tstart = atime;
				argv[i] = 0;
				break;

			case 'c':	/* Number of copies desired */
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "copies");
					error++;
					break;
				} else
					i++;
				if (copies != 0) {
					err (dupfmt, "copies");
					error++;
				} else if (checknum(argv[i],
				    &copies, Cp_min, Cp_max) == -1) {
				    	err(
				     "Copies must be in the range %d to %d",
					  Cp_min, Cp_max);
					error++;
				} else
					r.r_copies = copies;
				argv[i] = 0;
				break;

			case 'f':	/* What forms to print on */
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "forms");
					error++;
					break;
				} else
					i++;
				if (forms != 0) {
					err (dupfmt, "forms");
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "forms");
					error++;
				} else if ((forms = ckforms (argv[i])) == NULL) {
					err("Unknown form %s", argv[i]);
					error++;
				}
				argv[i] = 0;
				break;

			case 'h':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "header");
					error++;
					break;
				} else
					i++;
				if (header != 0) {
					err (dupfmt, "header");
					error++;
				} else if (access (argv[i], RACCESS) != 0) {
					err ("Cannot access header %s",
					  argv[i]);
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "header");
					error++;
				}
				header = argv[i];
				argv[i] = 0;
				break;

			case 'H':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "header");
					error++;
					break;
				} else
					i++;
				if (header != 0) {
					err (dupfmt, "header");
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "header");
					error++;
				} else {
					char hdrbuf[PATHSIZ];

					sprintf( hdrbuf, "%s/%s", Headerdir,
								argv[i] );
					if (access (hdrbuf, RACCESS) != 0) {
						err ("Cannot access header %s",
							hdrbuf );
						error++;
					}
					header = strdup(hdrbuf);
				}
				argv[i] = 0;
				break;

			case 'm':	/* Send mail upon completion */
				argv[i] = 0;
				qoptions |= O_MAIL;
				break;

			case  'n':	/* No copy, FF, indent */
				if (*++arg == 0) {
					err ("-n?, requires additional letter");
					error++;
				}
				switch (*arg) {
				case 'c':
					lpoptions |= NOCOPY;
					break;
				case 'f':
					lpoptions |= NOFF;
					break;
				case 'i':
					lpoptions |= NOINDENT;
					break;
				case 'k':
					lpoptions |= NOKEEP;
					break;
				default:
					err ("-n%c, unknown option", *arg);
					error++;
				}
				argv[i] = 0;
				break;

			case 'p':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "priority");
					error++;
					break;
				} else
					i++;
				if (priority != 0) {
					err (dupfmt, "priority");
					error++;
				} else if (checknum(argv[i],
				    &priority, Pri_min, Pri_max) == -1) {
					err(
				  "Priorities must be in the range %d to %d",
					  Pri_min, Pri_max);
					error++;
				} else
					r.r_pri = priority;
				argv[i] = 0;
				break;

			case 'q':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "queue");
					error++;
					break;
				} else
					i++;
				if (queue != 0) {
					err (dupfmt, "queue");
					error++;
				} else if((qp = findque( argv[i] )) == NULL) {
					err ("Unknown queue '%s'", argv[i]);
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "queue");
					error++;
				}
				queue = argv[i];
				argv[i] = 0;
				break;

			case 'u':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "user");
					error++;
					break;
				} else
					i++;
				if (username != 0) {
					err (dupfmt, "user");
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "user");
					error++;
				}
				username = argv[i];
				argv[i] = 0;
				break;

			case 'v':
				verbose++;
				argv[i] = 0;
				break;

			case 'x':
				argv[i] = 0;
				if (argv[i+1] == 0) {
					err (missingfmt, "extended");
					error++;
					break;
				} else
					i++;
				if (xargcnt == NXARGS) {
					err ("Too many extended arguments");
					error++;
				} else if (index (argv[i], '\n')) {
					err (newlinefmt, "extended");
					error++;
				} else
					xargs[xargcnt++] = argv[i];
				argv[i] = 0;
				break;

			case '?':
				error++;
				break;

			}
			arg++;
		}
	}

	if( error ) {
		fputs ("Usage:   qpr [options] files\n", stderr);
		fputs ("Options: [-a time] [-c #] [-f form] [-[hH] header] [-m]\n", stderr);
		fputs ("         [-n[cfik]] [-p pri] [-q que] [-t title] [-u useraddr] [-v] [-x arg]\n", stderr);
		exit (1);
	}

	if( queue == NULL )
		if((qp = findque( queue = Printqueue )) == NULL)
			bomb( "Default queue '%s' not configured\n", queue );

	if( forms == NULL ) {
		while (*++qp)
			if(strncmp(*qp, "form=", 5) == SAME) {
				forms = strdup(*qp+5);
				break;
			}
	}

	if( forms == NULL )
		forms = Printforms;

	if(( forms = ckforms( arg = forms )) == NULL )
		bomb( "Forms '%s' not configured\n", arg );

	umask (077);

	cf = mkcontrol ("QprXXXXXX", r.r_name);
	r.r_options = qoptions;
	strncpy (r.r_que, queue, QUESIZ);
	strncpy (r.r_forms, forms, FORMSIZ);

	/* Write out the request structure */
	puthdr (&r, cf);

	/* Write out the one line INTERNET address for notification */
	if( username == 0 ) {
		if ((pw = getpwuid (r.r_uid)) == NULL)
			fprintf( cf, "UUser Unknown <%s>\n", Sysmgr );
		else
			fprintf( cf, "U%s@%s\n", pw->pw_name, Hostname );
	} else
		fprintf( cf, "U%s\n", username );

	/* If special PRINT options were specified, note them in DECIMAL */
	if (lpoptions)
		fprintf (cf, "O%d\n", lpoptions);

	for (qp = &xargs[0]; xargcnt-- > 0; qp++)
		fprintf (cf, "X%s\n", *qp);

	if (lpoptions & NOCOPY)
		dumpdir( cf );



	/* Pass 2: Do the real work... */
	for (i = 1; i < argc; i++) {
		if (argv[i] == 0)
			continue;
		arg = argv[i];
		if (*arg != '-') {
			/* Must be a file... */
			nfiles++;
			process (argv[i]);
			oneheader = 0;	/* one header is enough when */
			if (title)		/* printing more than 1 file*/
				free (title);
			title = 0;
			continue;
		} else if (strcmp (arg, "-") == 0) {	/* DAG -- bug fix (was *arg) */
			/* Standard input specification
			nfiles++;
			dostdin();
			clearerr (stdin);	/* Important! */
			oneheader = 0;
			if (title)
				free (title);
			title = 0;
			continue;
		}
		arg++;
		while (*arg) {
			switch (*arg) {
			case 't':
				if (argv[i+1] == 0) {
					err (missingfmt, "title");
					error++;
					break;
				} else
					i++;
				if (title != 0) {
					free (title);
				}
				if (index (argv[i], '\n')) {
					err (newlinefmt, "title");
					error++;
				}
				title = strdup(argv[i]);
				break;
			}
			arg++;
		}
	}
	
	/*
	 *  SPECIAL CASE!
	 *
	 *  Read from the standard input, always create a datafile.
	 */
	if (nfiles == 0 && nfails == 0) {
		dostdin();
		fclose (cf);
		if( ferror( cf ) || ferror( df )) {
			err("Error writing queue files, aborting");
			exit( 99 );
		}
		queueit(&r, verbose);
		
		/*** Never reached ***/
	}
	
	fclose (cf);
	if( ferror( cf )) {
		err("Error writing queue files, aborting");
		exit( 97 );
	}
	if (nfiles && (nfails < nfiles))
		queueit(&r, verbose);
	else {
		printf("Nothing queued.\n");
		exit(0);
	}

	/** Never reached **/
}

dostdin()
{
	long	total;

	/* Process the standard input */
	df = mkdata (timenow++, dataname);
	total = copy (stdin, df, "standard input");
	fclose (df);
	if (title == NULL)
		fprintf (cf, "TSTANDARD-INPUT\n");
	else
		fprintf (cf, "T%s\n", title);
	/* if already did one header, don't do another */
	if (oneheader)   
		fprintf (cf, "H%s\n", header ? header : Printhdr);
	fprintf (cf, "C%ld\n", total);
	fprintf (cf, "I%s\n", dataname);
}

/*
 *	P R O C E S S (  )
 *
 *	Enqueue a single file.
 */
process (file)
char	*file;
{
	register char *cp;
	FILE *fin;

	if (access (file, RACCESS)) {
		perror (file);
		nfails++;
		err ("Cannot access %s, skipping...", file);
		return;
	}
#ifdef TRUNCATE
	if (cp = rindex(file, '/'))
		cp++;
	else
#endif TRUNCATE
		cp = file;

	if( lpoptions & NOCOPY ) {
		struct stat statb;

		fprintf( cf, "T%s\n", title ? title : cp);
		if (stat (file, &statb) == 0)
			fprintf( cf, "C%ld\n", statb.st_size);
		fprintf( cf, "F%s\n", file );
	}
	else if ((fin = fopen (file, "r")) == NULL) {
		perror (file);
		nfails++;
		err ("Cannot open %s, skipping...", file);
		return;
	} else {
		long	total;

		df = mkdata (timenow++, dataname);
		total = copy (fin, df, file);
		fclose (fin);
		fclose (df);
		if( ferror( df )) {
			err("Error writing queue files, aborting");
			exit( 98 );
		}
		fprintf( cf, "T%s\n", title ? title : cp);
		/* if already did one header, don't do another */
		if (oneheader)
			fprintf (cf, "H%s\n", header ? header : Printhdr);
		fprintf( cf, "C%ld\n", total);
		fprintf( cf, "I%s\n", dataname);
	}
	return;
}

cleanup(signal)
int	signal;
{
	rmdata( cf );
	if( r.r_name[0] ) {
		char	longname[PATHSIZ];
	
		sprintf (longname, "%s/%s", Tmpqdir, r.r_name);
		unlink (longname);
	}
	_exit (signal);
}
