/*
	vpl - MDQS despooler subroutine for UNIX plot on Versatec
			(using BRL non-standard Versatec driver)

	last edit:	31-Oct-1983	D A Gwyn


Function:
	Reads device-independent plot data from specified input file;
	for each frame, builds an image file containing raster data then
	sends the requested number of copies of this to the device,
	which is assumed to be already open on the standard output.

	Edge-limiting is done here; use "rot" if clipping is desired.

Method:
	Inputs vector data and builds a rasterization descriptor for
	each visible stroke.  (Strokes are limited to frame boundaries.)
	X goes down the page, Y goes from left to right.  To obtain a
	different orientation, pre-process data with the "rot" filter.

	The frame image file of SCANS scans is considered artificially
	divided into BANDS bands, each containing SPB scans.  Each band
	has a linked list of descriptors for not-yet-rasterized strokes
	that start in the band.

	Space for descriptors is obtained via "malloc".  When no more
	space is available, the image file is updated as follows, then
	"malloc" is tried again ("must" work the second time):

	Each band in increasing X order becomes "active"; if no
	descriptors exist for the band it is skipped, otherwise its
	existing raster data is re-read from the image file into a
	buffer and each descriptor is processed to rasterize its stroke.
	If the stroke terminates in the band its descriptor is freed,
	otherwise the descriptor is linked into the following band's
	list.  When the descriptor list for the active band becomes
	empty (must happen), the band's raster data is flushed back to
	the image file and the next band becomes active.  This process
	continues until all vectors have been input and rasterized.

Acknowledgment:
	Based very loosely on Mike Muuss's Versatec TIGpack interpreter.
*/


#include	<signal.h>
#include	<stdio.h>

/*
	Raster device model and image terminology as used herein:

Frames are delimited in plot data by "erase" codes; each frame looks
like:


			    top of frame
	---------------------------------------------------------
	|-----> plot						|
	||	Y axis						|
 ^	||							|
 |	|v		Plot is "sideways" on paper.		|
	|plot							|
paper	 X axis
feed			       nib
	|			|				|
 ^	|			v				|
 |	|			.				|
	|scan line ->...........................................|
	|			.				|
	---------------------------------------------------------
			    bottom of frame


Each plot-mode scan line consists of exactly BYTES bytes of plot data.

A single 8-bit output byte of plot data represents BITS (== 8) nibs,
most significant bit leftmost, with a 1 bit marking and 0 bit clear.
Successive plot data bytes produce the image from left to right.
*/


/*	Versatec D1200A and V80 device parameters:	*/

#define NIBS	(132 * 16)		/* nibs per scan */
#define SCANS	(2 * 68 * 25)		/* scans chosen per frame */
#define NPI	200			/* nibs/inch across page */
#define SPI	200			/* scans/inch down page */

#define BITS	8			/* nibs per output byte */
#define BitSet( i )	(0200 >> (i))	/* bit to set in plot byte */
#define CLEAR	0			/* plot code with nibs off */

/* the following parameter should be tweaked for fine-tuning */
#define SPB	10			/* scans per band */

#if	SCANS % SPB != 0
#include	"SCANS % SPB != 0"	/* (just for ERROR message) */
#endif

/* the following parameter is used only for user line limit quota */
#define LPP	(2 * 68L)		/* text_lines/frame conversion */

/*	Program constants computed from device parameters:	*/

#define BANDS	(SCANS / SPB)		/* # of "bands" */
#define BYTES	(NIBS / BITS)		/* max data bytes per scan */
#define XMAX	(SCANS - 1)		/* frame boundary */
#define YMAX	(NIBS - 1)		/* scan boundary */
/* The following assumes NPI is an integral multiple of SPI;
   a more general algorithm would have to use long integers. */
#if	NIBS / NPI < SCANS / SPI	/* (integer approximation) */
#define XSCA	(NIBS / (NPI / SPI))
#define YSCA	NIBS
#else
#define XSCA	SCANS
#define YSCA	(SCANS * (NPI / SPI))
#endif


/*	Data structure definitions:	*/

#ifdef pdp11
typedef char	tiny;			/* for very small numbers */
#else
typedef	short	tiny;			/* for very small numbers */
#endif

typedef int	bool;			/* boolean data type */
#define false	0
#define true	1

typedef struct
	{
	short		x;
	short		y;
	}	coords; 		/* Cartesian coordinates */

typedef struct descr
	{
	struct descr	*next;		/* next in list, or NULL */
	coords		pixel;		/* starting scan, nib */
	tiny		xsign;		/* 0 or +1 */
	tiny		ysign;		/* -1, 0, or +1 */
	bool		xmajor; 	/* true iff X is major dir. */
	short		major;		/* major dir delta (nonneg) */
	short		minor;		/* minor dir delta (nonneg) */
	short		e;		/* DDA error accumulator */
	short		de;		/* increment for `e' */
	}	stroke; 		/* rasterization descriptor */

/*	Global data allocations:	*/

static struct
	{
	short		left;		/* window edges */
	short		bottom;
	short		right;
	short		top;
	}	space =	 		/* plot scale data */
	{ 0, 0, XSCA, YSCA };		/* default unity scaling */

static long	delta = YSCA;		/* larger window dimension */
static long	deltao2 = YSCA / 2;	/* delta / 2 */

static stroke	*band[BANDS] = { NULL };	/* descriptor lists */

/* active band buffer: */
static char	buffer[SPB][BYTES] = { 0 };
static short	xstart = 0;		/* active band starting scan */

static char	iname[] = "/usr/tmp/vpl_io";
static int	ifd = -1;		/* for image file I/O */

static FILE	*pfp = NULL;		/* for plot file input */

static int	sigs[] =		/* signals to be caught */
	{
	SIGHUP,
	SIGINT,
	SIGQUIT,
	SIGTERM
	};


/*	Externals:	*/

extern void	free();
extern char	*malloc();
extern int	close(), creat(), getpid(), kill(), open(), read(),
		write();
extern long	lseek();


/*	Local subroutines:	*/

static int	DoFile(), Foo();
static stroke	*Allocate(), *Dequeue();
static bool	BackBand(), BuildStr(), CopyOut(), GetCoords(),
		InitFile(), OutBuild(), RdBand(), Rescan(), WrBand();
static void	Catch(), FreeUp(), InitDesc(), PutDot(), Queue(),
		Raster(), SetSigs();

/*
	vpl - rasterizer control subroutine

	This routine reads UNIX plot records from the specified file
	and controls the entry of the strokes into the descriptor lists.
	Strokes are limited (not clipped) to fit the frame.

	Upon end of file or erase, plot data is copied to the device.
*/

int
vpl( copies, pname, llimit, plused )	/* returns status code:
					   < 0	=> catastrophe
					   = 0	=> complete success
					   > 0	=> line limit hit   */
	int	copies; 		/* # copies of each frame */
	char	*pname; 		/* name of plot data file */
	long	llimit; 		/* remaining user line limit */
	long	*plused;		/* receives actual lines used */
	{
	if ( plused == NULL )
		return -1;		/* sick caller! */
	*plused = 0L;			/* no lines used yet */

	if ( pname == NULL )
		return -2;		/* sick caller! */

	if ( llimit <= 0L )
		return 1;		/* lazy caller! */

	if ( copies <= 0 )
		copies = 1;		/* dumb caller! */

	SetSigs();			/* set signal catchers */

	/* open raster image file and UNIX plot input file */

	if ( (ifd = open( iname, 2 )) < 0
	  && ( close( creat( iname, 0600 ) ) < 0
	    || (ifd = open( iname, 2 )) < 0
	     )
	   )
		return Foo( -3 );	/* can't open image file */

	if ( (pfp = fopen( pname, "r" )) == NULL )
		return Foo( -4 );	/* file not found */

	return DoFile( copies, llimit, plused );	/* plot it */
	}

/*
	DoFile - process UNIX plot file onto Versatec
*/

static int
DoFile( copies, llimit, plused )	/* returns vpl status code */
	int		copies; 	/* # copies of each frame */
	long		llimit; 	/* remaining user line limit */
	long		*plused;	/* receives actual lines used */
	{
	register bool	plotted;	/* false => empty frame image */
	register int	c;		/* input character */
	coords		newpos; 	/* current input coordinates */
	coords		virpos; 	/* virtual pen position */

	/* process each frame into a raster image file */

	for ( ; ; )			/* for each frame */
		{
		if ( !InitFile() )	/* create empty frame image */
			return Foo( -5 );	/* can't create oname */

		InitDesc();		/* empty descriptor lists */

		virpos.x = virpos.y = 0;
		plotted = false;

		for ( ; ; )		/* read until EOF or erase */
			{
			switch ( c = getc( pfp ) )
				{	/* record type */
			case EOF:
			case 'e':	/* erase */
				if ( plotted )
					{
					/* flush strokes */

					if ( !OutBuild() )
						return Foo( -6 );

					/* dump multiple copies */
					{
					int	copy;

					copy = copies;
					do	{
						if ( !CopyOut() )
							return Foo( -7 );

						if ( (*plused += LPP)
								> llimit
						   )
							return Foo( 2 );
						}
					while ( --copy != 0 );
					}
					}

				if ( c == EOF )
					return Foo( 0 );/* success */

				break;	/* next frame */

			case 'f':	/* linemod */
				/* ignore for time being */
				while ( (c = getchar()) != EOF
				     && c != '\n'
				      )
					;	/* eat string */
				continue;

			case 'l':	/* line */
			case 'm':	/* move */
				if ( !GetCoords( &newpos ) )
					return Foo( -8 );
				virpos = newpos;
				if ( c != 'l' )
					continue;
				/* line: fall through */

			case 'n':	/* cont */
			case 'p':	/* point */
				if ( !GetCoords( &newpos ) )
					return Foo( -9 );
				if ( c == 'p' )
					virpos = newpos;
				if ( !BuildStr( &virpos, &newpos ) )
					return Foo( -10 );
				plotted = true;
				virpos = newpos;
				continue;

			case 's':	/* space */
				if ( fread( (char *)&space,
					    (int)sizeof space, 1, pfp
					  ) != 1
				   )
					return Foo( -11 );

				delta = (long)space.right
				      - (long)space.left;
				deltao2 = (long)space.top
					- (long)space.bottom;
				if ( deltao2 > delta )
					delta = deltao2;
				deltao2 = (delta + 1L) / 2L;
				continue;

			case 't':	/* label */
				/* ignore for time being */
				while ( (c = getchar()) != EOF
				     && c != '\n'
				      )
					;	/* eat string */
				continue;

			default:
				return Foo( -12 );	/* bad input */
				}
			break;
			}		/* next input record */
		}			/* next frame */
	}

/*
	GetCoords - input x,y coordinates and scale into pixels
*/

static bool
GetCoords( coop )
	register coords	*coop;		/* -> input coordinates */
	{
	/* read coordinates */

	if ( fread( (char *)coop, (int)sizeof (coords), 1, pfp ) != 1 )
		return false;

	/* limit left, bottom */

	if ( (coop->x -= space.left) < 0 )
		coop->x = 0;
	if ( (coop->y -= space.bottom) < 0 )
		coop->y = 0;

	/* convert to device pixels */

	coop->x = (short)(((long)coop->x * XSCA + deltao2) / delta);
	coop->y = (short)(((long)coop->y * YSCA + deltao2) / delta);

	/* limit right, top */

	if ( coop->x > XMAX )
		coop->x = XMAX;
	if ( coop->y > YMAX )
		coop->y = YMAX;

	return true;
	}

/*
	CopyOut - copy final image file to Versatec one scan at a time
*/

static bool
CopyOut()				/* returns true if successful */
	{
	register short	scan;		/* counts scans copied */

	if ( !Rescan() )		/* beginning of image file */
		return false;

	scan = SCANS;
	do
		if ( read( ifd, &buffer[0][0], sizeof buffer[0] )
						!= sizeof buffer[0]
		  || write( fileno( stdout ), &buffer[0][0],
			    sizeof buffer[0]
			  ) != sizeof buffer[0]
		   )
			return false;
	while ( --scan != 0 );

	return true;			/* success */
	}

/*
	InitDesc - initialize stroke descriptor lists
*/

static void
InitDesc()
	{
	register stroke **bp;		/* *bp -> start of descr list */

	for ( bp = &band[0]; bp < &band[BANDS]; ++bp )
		*bp = NULL;		/* nothing in band yet */
	}


/*
	Queue - queue descriptor onto band list

	Note that descriptor order is not important.
*/

static void
Queue( hp, vp )
	register stroke **hp;		/* *hp -> first descr in list */
	register stroke *vp;		/* -> new descriptor */
	{
	vp->next = *hp; 		/* -> first in existing list */
	*hp = vp;			/* -> new first descriptor */
	}


/*
	Dequeue - remove descriptor from band list (do not free space)
*/

static stroke *
Dequeue( hp )				/* returns addr of descriptor,
					   or NULL if none left */
	register stroke **hp;		/* *hp -> first descr in list */
	{
	register stroke *vp;		/* -> descriptor */

	if ( (vp = *hp) != NULL )
		*hp = vp->next; 	/* -> next element in list */

	return vp;			/* may be NULL */
	}


/*
	FreeUp - deallocate descriptors
*/

static void
FreeUp()
	{
	register stroke **hp;		/* *hp -> head of descr list */
	register stroke *vp;		/* -> rasterization descr */

	for ( hp = &band[0]; hp < &band[BANDS]; ++hp )
		while ( (vp = Dequeue( hp )) != NULL )
			free( (char *)vp );	/* free storage */
	}

/*
	InitFile - initialize frame raster image file
*/

static bool
InitFile()				/* returns true iff successful */
	{
	register char	*rp;		/* -> buffer[.][.] */
	register short	b;		/* band index */

	/* create an empty band image in active-band buffer: */

	for ( rp = &buffer[0][0]; rp < &buffer[SPB][0]; )
		*rp++ = CLEAR;		/* all nibs off */

	/* write all (empty) bands to image file: */

	if ( !Rescan() )		/* beginning of image file */
		return false;

	b = BANDS;
	do
		if ( !WrBand() )
			return false;	/* can't write image file */
	while ( --b != 0 );

	return true;
	}


/*
	Rescan - reposition raster image file at beginning
*/

static bool
Rescan()				/* returns true if successful */
	{
	return lseek( ifd, 0L, 0 ) == 0L;
	}


/*
	BackBand - reposition raster image file backward one band
*/

static bool
BackBand()				/* returns true if successful */
	{
	return lseek( ifd, -(long)sizeof buffer, 1 ) >= 0L;
	}


/*
	RdBand - Read in next band from input image file
*/

static bool
RdBand()
	{
	return read( ifd, &buffer[0][0], sizeof buffer ) ==
						(int)sizeof buffer;
	}


/*
	PutDot - set pixel in active-band buffer
*/

static void
PutDot( relx, absy )
	register short	relx;		/* relative to `xstart' */
	register short	absy;
	{
	buffer[relx][absy / BITS] |= BitSet( absy % BITS );
	}


/*
	WrBand - Write out next band to output image file
*/

static bool
WrBand()
	{
	return write( ifd, &buffer[0][0], sizeof buffer ) ==
						(int)sizeof buffer;
	}

/*
	BuildStr - set up DDA parameters and queue stroke
*/

static bool
BuildStr( pt1, pt2 )			/* returns true unless bug */
	coords		*pt1, *pt2;	/* endpoints */
	{
	register stroke *vp;		/* -> rasterization descr */

	/* arrange for pt1 to have the smaller X-coordinate: */

	if ( pt1->x > pt2->x )
		{
		register coords *temp;	/* temporary for swap */

		temp = pt1;		/* swap pointers */
		pt1 = pt2;
		pt2 = temp;
		}

	if ( (vp = Allocate()) == NULL )	/* alloc a descriptor */
		return false;		/* "can't happen" */

	/* set up multi-band DDA parameters for stroke */

	vp->pixel = *pt1;		/* initial pixel */
	vp->major = pt2->x - vp->pixel.x;	/* always nonnegative */
	vp->xsign = vp->major ? 1 : 0;
	vp->minor = pt2->y - vp->pixel.y;
	if ( (vp->ysign = vp->minor ? (vp->minor > 0 ? 1 : -1) : 0) < 0
	   )
		vp->minor = -vp->minor;

	/* if X is not really major, correct the assignments */

	if ( !(vp->xmajor = vp->minor <= vp->major) )
		{
		register short	temp;	/* temporary for swap */

		temp = vp->minor;
		vp->minor = vp->major;
		vp->major = temp;
		}

	vp->e = vp->major / 2 - vp->minor;	/* initial DDA error */
	vp->de = vp->major - vp->minor;

	/* link descriptor into band corresponding to starting scan */

	Queue( &band[vp->pixel.x / SPB], vp );

	return true;
	}

/*
	Allocate - allocate a descriptor (possibly updating image file)
*/

static stroke *
Allocate()				/* returns addr of descriptor
					   or NULL if something wrong */
	{
	register stroke *vp;		/* -> allocated storage */

	if ( (vp = (stroke *)malloc( sizeof(stroke) )) == NULL
	  && OutBuild() 		/* flush and free up storage */
	   )
		vp = (stroke *)malloc( sizeof(stroke) );

	return vp;			/* should be non-NULL now */
	}

/*
	OutBuild - rasterize all strokes into raster frame image
*/

static bool
OutBuild()				/* returns true if successful */
	{
	register stroke **hp;		/* *hp -> head of descr list */
	register stroke *vp;		/* -> rasterization descr */
	register stroke **np;		/* `hp' for next band */

	for ( hp = &band[0]; hp < &band[BANDS]; ++hp )
		if ( *hp != NULL )
			break;
	if ( hp == &band[BANDS] )
		return true;		/* nothing to do */

	if ( !Rescan() )		/* beginning of image file */
		return false;

	for ( hp = &band[0], np = &band[1], xstart = 0;
	      hp < &band[BANDS];
	      hp = np++, xstart += SPB
	    )	{
		/* roll next band's image into active-band buffer */

		if ( !RdBand() )
			return false;	/* can't read image file */

		while ( (vp = Dequeue( hp )) != NULL )
			Raster( vp, np );      /* rasterize stroke */
		/* Raster() either re-queued the descriptor onto the
		   next band list or else it freed the descriptor */

		if ( !BackBand() || !WrBand() ) /* replace band image */
			return false;	/* can't write image file */
		}

	return true;			/* success */
	}

/*
	Raster - rasterize stroke.  If overflow into next band, requeue;
					else free the descriptor.

Method:
	Modified Bresenham algorithm.  Guaranteed to mark a dot for
	a zero-length stroke.  Please do not try to "improve" this code
	as it is extremely hard to get all aspects just right.
*/

static void
Raster( vp, np )
	register stroke *vp;		/* -> rasterization descr */
	stroke		**np;		/* *np -> next band 1st descr */
	{
	register short	dx;		/* raster within active band */

	for ( dx = vp->pixel.x - xstart; dx < SPB; )
		{
		PutDot( dx, vp->pixel.y );	/* ensure one dot min */
		if ( vp->major-- == 0 ) /* done! */
			{
			free( (char *)vp );	/* return to "malloc" */
			return;
			}
		if ( vp->e < 0 )	/* advance major & minor */
			{
			dx += vp->xsign;
			vp->pixel.y += vp->ysign;
			vp->e += vp->de;
			}
		else	{		/* advance major only */
			if ( vp->xmajor )	/* X is major dir */
				++dx;
			else			/* Y is major dir */
				vp->pixel.y += vp->ysign;
			vp->e -= vp->minor;
			}
		}

	/* overflow into next band; re-queue */

	vp->pixel.x = xstart + SPB;
	Queue( np, vp );	       /* DDA parameters already set */
	}

/*
	Foo - clean up before return from rasterizer
*/

static int
Foo( code )				/* returns status code */
	int	code;			/* status code */
	{
	(void)close( ifd );		/* close open files */
	if ( pfp != NULL )
		(void)fclose( pfp );

	(void)unlink( iname );		/* remove raster image file */

	FreeUp();			/* deallocate descriptors */

	return code;
	}

/*
	SetSigs - set up signal catchers
*/
static void
SetSigs()
	{
	register int	*psig;		/* -> sigs[.] */

	for ( psig = &sigs[0];
	      psig < &sigs[sizeof sigs / sizeof sigs[0]];
	      ++psig
	    )
		if ( signal( *psig, SIG_IGN ) != SIG_IGN )
			(void)signal( *psig, Catch );
	}


/*
	Catch - invoked on interrupt
*/

static void
Catch( sig )
	register int	sig;		/* signal number */
	{
	register int	pid;		/* this process's ID */
	register int	*psig;		/* -> sigs[.] */

	for ( psig = &sigs[0];
	      psig < &sigs[sizeof sigs / sizeof sigs[0]];
	      ++psig
	    )
		(void)signal( *psig, SIG_IGN );

	(void)Foo( -13 );		/* clean up */

	(void)signal( sig, SIG_DFL );

	if ( (pid = getpid()) > 1 )
		(void)kill( pid, sig ); /* resignal process */
	}
